@@ -1,4 +1,5 @@
 import time
+import datetime
 import logging
 import paho.mqtt.client as mqtt
 from google.protobuf.json_format import MessageToDict
@@ -34,6 +35,8 @@ class MqttSpbTopic:
 
         self.entity_name = None
 
+        self.domain = None
+
         if topic_str is not None:
             self.parse_topic(topic_str)
 
@@ -67,6 +70,10 @@ class MqttSpbTopic:
             self.eon_device_name = topic_fields[4]
             self.entity_name = self.eon_device_name
 
+        self.domain = "%s.%s.%s" % ( self.namespace, self.group_name, self.eon_name)
+        if self.eon_device_name is not None:
+            self.domain += ".%s" % self.eon_device_name
+
         return str(self)
 
 
@@ -75,13 +82,13 @@ class MqttSpbPayload:
         Class to parse binary payloads into dictionary
     """
 
-    def __init__(self, payload_data = None):
+    def __init__(self, payload_data=None):
 
         self.payload = None
 
         # If data is passed, then process it
         if payload_data is not None:
-            return self.parse_payload(payload_data)
+            self.parse_payload(payload_data)
 
     def __str__(self):
         return str(self.payload)
@@ -112,7 +119,12 @@ class MqttSpbPayload:
         except Exception as e:
 
             # Check if payload is from SCADA
-            _payload = payload_data.decode()
+            try:
+                _payload = payload_data.decode()
+            except Exception as e1:
+                logger.error("Could not parse MQTT CMD payload, message ignored ! (reason: %s)" % (str(e1)))
+                return None
+
             if _payload == "OFFLINE" or _payload == "ONLINE":
                 self.payload = _payload
                 return self.payload
@@ -128,7 +140,7 @@ class MqttSpbPayload:
 class MqttSpbEntity:
 
     def __init__(self, spb_group_name, spb_eon_name,
-                 spb_eon_device_name = None,
+                 spb_eon_device_name= None,
                  debug_info=False,
                  filter_cmd_msg=True,
                  entity_is_scada=False):
@@ -146,6 +158,7 @@ class MqttSpbEntity:
         self.entity_subclass = None
 
         self.is_alive = True
+        self.is_birth_published = False
 
         self.attribures = self._ValuesGroup()
         self.data = self._ValuesGroup()
@@ -153,6 +166,7 @@ class MqttSpbEntity:
 
         self.on_command = None  # Callback function when a comand is received
         self.on_connect = None
+        self.on_disconnect = None
         self.on_message = None
 
         # Private members -----------
@@ -162,9 +176,9 @@ class MqttSpbEntity:
         self._spb_eon_device_name = spb_eon_device_name
 
         if spb_eon_device_name is None:
-            self._entity_domain = "%s.%s" % (self._spb_group_name, self._spb_eon_name)
+            self._entity_domain = "spBv.10.%s.%s" % (self._spb_group_name, self._spb_eon_name)
         else:
-            self._entity_domain = "%s.%s.%s" % (self._spb_group_name, self._spb_eon_name, self._spb_eon_device_name)
+            self._entity_domain = "spBv.10.%s.%s.%s" % (self._spb_group_name, self._spb_eon_name, self._spb_eon_device_name)
 
         if spb_eon_device_name is None:
             self._entity_name = self._spb_eon_name
@@ -237,13 +251,96 @@ class MqttSpbEntity:
 
         return MetricDataType.Unknown
 
+    def serialize_payload_birth(self):
+        """
+            Serialize the BIRTH message and get payload bytes
+        """
+
+        if self._spb_eon_device_name is None:  # EoN type
+            payload = getNodeBirthPayload()
+        else:  # Device
+            payload = getDeviceBirthPayload()
+
+        # Attributes
+        if not self.attribures.is_empty():
+            for item in self.attribures.values:
+                name = "ATTR/" + item.name
+                addMetric(payload, name, None, self._spb_data_type(item.value), item.value, item.timestamp)
+
+        # Data
+        if not self.data.is_empty():
+            for item in self.data.values:
+                name = "DATA/" + item.name
+                addMetric(payload, name, None, self._spb_data_type(item.value), item.value, item.timestamp)
+
+        # Commands
+        if not self.commands.is_empty():
+            for item in self.commands.values:
+                name = "CMD/" + item.name
+                addMetric(payload, name, None, self._spb_data_type(item.value), item.value, item.timestamp)
+
+        payload_bytes = bytearray(payload.SerializeToString())
+
+        return payload_bytes
+
+    def deserialize_payload_birth(self, data_bytes):
+
+        payload = MqttSpbPayload(data_bytes).payload
+
+        if payload is not None:
+
+            # Iterate over the metrics to update the data fields
+            for field in payload.get('metrics', []):
+
+                if field['name'].startswith("ATTR/"):
+                    field['name'] = field['name'][5:]
+                    self.attribures.set_value(field['name'], field['value'], field['timestamp'])  # update field
+
+                elif field['name'].startswith("CMD/"):
+                    field['name'] = field['name'][4:]
+                    self.commands.set_value(field['name'], field['value'], field['timestamp'])  # update field
+
+                elif field['name'].startswith("DATA/"):
+                    field['name'] = field['name'][5:]
+                    self.data.set_value(field['name'], field['value'], field['timestamp'])  # update field
+
+        return payload
+
+    def serialize_payload_data(self, send_all=False):
+
+        # Get a new payload object to add metrics to it.
+        payload = getDdataPayload()
+
+        # Iterate for each data field.
+        for item in self.data.values:
+
+            # Only send those values that have been updated, or if send_all==True then send all.
+            if send_all or item.is_updated:
+                addMetric(payload, item.name, None, self._spb_data_type(item.value), item.value, item.timestamp)
+
+        payload_bytes = bytearray(payload.SerializeToString())
+
+        return payload_bytes
+
+    def deserialize_payload_data(self, data_bytes):
+
+        payload = MqttSpbPayload(data_bytes).payload
+
+        if payload is not None:
+
+            #Iterate over the metrics to update the data fields
+            for field in payload.get('metrics', []):
+                self.data.set_value(field['name'], field['value'], field['timestamp']) # update field
+
+        return payload
+
     def publish_birth(self):
 
         if not self.is_connected():  # If not connected
             logger.warning("%s - Could not send publish_birth(), not connected to MQTT server" % self._entity_domain)
             return False
 
-        if self.is_empty():  # If no data (Telemetry, attributes, commands )
+        if self.is_empty():  # If no data (Data, attributes, commands )
             logger.warning(
                 "%s - Could not send publish_birth(), entity doesn't have data ( attributes, data, commands )" % self._entity_domain)
             return False
@@ -256,32 +353,8 @@ class MqttSpbEntity:
             logger.info("%s - Published STATE BIRTH message " % (self._entity_domain))
             return
 
-        # PAYLOAD
-        if self._spb_eon_device_name is None:  # EoN type
-            payload = getNodeBirthPayload()
-        else:  # Device
-            payload = getDeviceBirthPayload()
-
-        # Attributes
-        if not self.attribures.is_empty():
-            for item in self.attribures.values:
-                name = "ATTR/" + item.name
-                addMetric(payload, name, None, self._spb_data_type(item.value), item.value)
-
-        # Telemetry
-        if not self.data.is_empty():
-            for item in self.data.values:
-                name = "DATA/" + item.name
-                addMetric(payload, name, None, self._spb_data_type(item.value), item.value)
-
-        # Commands
-        if not self.commands.is_empty():
-            for item in self.commands.values:
-                name = "CMD/" + item.name
-                addMetric(payload, name, None, self._spb_data_type(item.value), item.value)
-
         # Publish BIRTH message
-        payload_bytes = bytearray(payload.SerializeToString())
+        payload_bytes = self.serialize_payload_birth()
         if self._spb_eon_device_name is None:  # EoN
             topic = "spBv1.0/" + self.spb_group_name + "/NBIRTH/" + self._spb_eon_name
         else:
@@ -291,6 +364,8 @@ class MqttSpbEntity:
 
         logger.info("%s - Published BIRTH message" % (self._entity_domain))
 
+        self.is_birth_published = True
+
     def publish_data(self, send_all=False):
         """
             Send the new updated data to the MQTT broker as a Sparkplug B DATA message.
@@ -304,22 +379,16 @@ class MqttSpbEntity:
                 "%s - Could not send publish_telemetry(), not connected to MQTT server" % self._entity_domain)
             return False
 
-        if self.is_empty():  # If no data (Telemetry, attributes, commands )
+        if self.is_empty():  # If no data (Data, attributes, commands )
             logger.warning(
                 "%s - Could not send publish_telemetry(), entity doesn't have data ( attributes, data, commands )" % self._entity_domain)
             return False
 
-        # PAYLOAD - Telemetry Only
-        payload = getDdataPayload()
+        # Send payload if there is new data or we need to send all
+        if send_all or self.data.is_updated():
 
-        for item in self.data.values:
-            # Only send those values that have been updated, or if send_all==True then send all.
-            if send_all or item.is_updated:
-                addMetric(payload, item.name, None, self._spb_data_type(item.value), item.value)
+            payload_bytes = self.serialize_payload_data(send_all)   # Get the data payload
 
-        # Send payload if there is new data
-        if payload.metrics:
-            payload_bytes = bytearray(payload.SerializeToString())
             if self._spb_eon_device_name is None:  # EoN
                 topic = "spBv1.0/" + self.spb_group_name + "/NDATA/" + self._spb_eon_name
             else:
@@ -333,11 +402,20 @@ class MqttSpbEntity:
         logger.warning("%s - Could not publish DATA message, may be data no new data values?" % (self._entity_domain))
         return False
 
-    def connect(self, host='localhost', port=1883, user = "", password = ""):
+    def connect(self,
+                host='localhost',
+                port=1883,
+                user="",
+                password="",
+				use_tls=False,
+                tls_ca_path="",
+                tls_cert_path="",
+                tls_key_path="",
+                timeout=5):
 
         # If we are already connected, then exit
         if self.is_connected():
-            return
+            return True
 
         # MQTT Client configuration
         if self._mqtt is None:
@@ -350,6 +428,23 @@ class MqttSpbEntity:
         if user != "":
             self._mqtt.username_pw_set(user, password)
 
+        # If client certificates are provided
+        if tls_cert_path and tls_cert_path and tls_key_path:
+            logger.debug("Setting CA client certificates")
+            import ssl
+            self._mqtt.tls_set(ca_certs=tls_cert_path, certfile=tls_cert_path, keyfile=tls_key_path, cert_reqs=ssl.CERT_NONE)
+            # self._mqtt.tls_insecure_set(True)
+
+        #If only CA is proviced.
+        elif tls_cert_path:
+            logger.debug("Setting CA certificate")
+            self._mqtt.tls_set(ca_certs=tls_ca_path)
+
+        # If TLS is enabled
+        else:
+            if use_tls:
+                self._mqtt.tls_set()    # Enable TLS encryption
+
         # Entity DEATH message - last will message
         if self._entity_is_scada:  # If it is a type entity SCADA, change the DEATH certificate
             topic = "spBv1.0/" + self.spb_group_name + "/STATE/" + self._spb_eon_name
@@ -371,10 +466,16 @@ class MqttSpbEntity:
             logger.warning("%s - Could not connect to MQTT server (%s)" % (self._entity_domain, str(e)))
             return False
 
-        time.sleep(0.1)
         self._mqtt.loop_start()  # Start MQTT background task
+        time.sleep(0.1)
 
-        return True
+        # Wait some time to get connected
+        _timeout = time.time() + timeout
+        while not self.is_connected() and _timeout > time.time():
+            time.sleep(0.1)
+
+        # Return if we connected successfully
+        return self.is_connected()
 
     def disconnect(self):
 
@@ -412,9 +513,6 @@ class MqttSpbEntity:
             self._mqtt.subscribe(topic)
             logger.info("%s - Subscribed to MQTT topic: %s" % (self._entity_domain, topic))
 
-            # Publish Birth data
-            self.publish_birth()
-
         else:
             logger.error(" %s - Could not connect to MQTT server !" % self._entity_domain)
 
@@ -425,13 +523,17 @@ class MqttSpbEntity:
     def _mqtt_on_disconnect(self, client, userdata, rc):
         logger.info("%s - Disconnected from MQTT server" % self._entity_domain)
 
+        # Execute the callback function if it is not None
+        if self.on_disconnect is not None:
+            self.on_disconnect(rc)
+
     def _mqtt_on_message(self, client, userdata, msg):
 
         # Check if loopback message
         if self._loopback_topic == msg.topic:
                 return
 
-        msg_ts_rx = int(round(time.time() * 1000))  # Save the current timestamp
+        msg_ts_rx = int(round(datetime.datetime.now().utcnow().time() * 1000))  # Save the current timestamp
 
         logger.info("%s - Message received  %s" % (self._entity_domain, msg.topic))
 
@@ -510,7 +612,7 @@ class MqttSpbEntity:
             self.name = name
             self._value = value
             if timestamp is None:
-                self._timestamp = int(time.time() * 1000)
+                self._timestamp = int(datetime.datetime.now().utcnow().time() * 1000)
             else:
                 self._timestamp = timestamp
             self.is_updated = True
@@ -535,8 +637,9 @@ class MqttSpbEntity:
 
         @value.setter
         def value(self, value):
-            if value != self._value:
-                self.is_updated = True
+            # if value != self._value:
+            #     self.is_updated = True
+            self.is_updated = True
             self._value = value
 
         @property
@@ -551,7 +654,7 @@ class MqttSpbEntity:
                 self._timestamp = int(value)
 
         def timestamp_update(self):
-            self.timestamp = int(time.time() * 1000)
+            self.timestamp = int(datetime.datetime.utcnow().timestamp() * 1000)
 
     class _ValuesGroup:
 
@@ -613,7 +716,7 @@ class MqttSpbEntity:
             for item in self.values:
                 if item.name == name:
                     item.value = value
-                    item.timestamp = timestamp
+                    item.timestamp = timestamp  # If timestamp is none, the current time will be used.
                     return True
 
             # item was not found, then add it to the list.
@@ -621,16 +724,17 @@ class MqttSpbEntity:
 
             return True
 
-        def set_dictionary(self, values: dict):
+        def set_dictionary(self, values: dict, timestamp=None):
             """
                 Import a list of values based on a dictionary FieldName:FieldValue
             :param values:      Dictionary with fields-values
+            :param timestamp:   Timestamp value in ms
             :return:            Result
             """
 
             # Update the list of values
             for k,v in values.items():
-                self.set_value(k, v)
+                self.set_value(k, v, timestamp)
 
             return True
 
@@ -663,7 +767,7 @@ class MqttSpbEntityEdgeNode(MqttSpbEntity):
                 "%s - Could not send publish_command_device(), commands not provided or not valid. Please provide a dictionary of command:value" % self._entity_domain)
             return False
 
-        # PAYLOAD
+        # Get a new payload object, to add metrics
         payload = getDdataPayload()
 
         # Add the list of commands to the payload metrics
